/*	evt_queue.h	*/

#ifndef _EVT_QUEUE_H_
#define _EVT_QUEUE_H_

/*
 * Copyright 2007-2016 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

/*
 * The UNH IOL implementation of the Extended Sockets API (ES-API) Issue 1.0
 *
 * defined in:	The Interconnect Software Consortium Issue 1.0
 *		ISBN: 1-931624-52-6
 *		Published by The open Group, January 2005.
 *
 * all page references in this source code are to pages in that document.
 */

/*
 * Most of the coding in this file done by:
 *	Bob Russell
 *	rdr@iol.unh.edu
 */


#include <exs.h>
#include <exs_private.h>
#include <list.h>


/* set to non-zero to enable debug printing */
extern int _our_debug_flag;


/* flags for use with _cnv_qnumb_to_ptr */
enum ref_flags {
	EVT_QUEUE_REF_PENDING = 1,
};


/* Parameters for pending connects, but present in all connections */
struct pending_connect {
	struct evt_queue	*qhandle;
	exs_ahandle_t		ahandle;
	int			err_numb;
	int			sockno;
	int			evt_type;
	int			old_state;
};


struct evt_queue
{
	/* the number of active references to this queue */
	exs_atomic_uint_t	refcount;

	/* the number of pending events on this queue */
	exs_atomic_uint_t	npending;

	/* the list of full "struct exs_event" objects on this queue */
	struct our_list_head	full_list;

	/* the list of empty "struct exs_event" objects on this queue */
	struct our_list_head	empty_list;

	/* added for polling */
	unsigned long		qnumb;

	/* if > 0, the enabled signal
	 * if <= 0, signalling is disabled
	 */
	int			signal_number;

	/* the number of "struct exs_event" objects allocated to this queue */
	unsigned int		nitems;

	/* the current depth value defined by user */
	unsigned int		depth;

	/* 0 if using busy poll, 1 if using wait-wakeup */
	unsigned int		wait;

	/* maximum number of times to spin if using EXS_WAIT_ADAPTIVE */
	unsigned long long	max_spins;

	/* For performance testing */
#if defined(ALLOW_EXS_STATISTICS)
	unsigned long		n_waits;
	unsigned long		min_spin_count[20];
	unsigned long		max_spin_count[20];
	unsigned long 		spin_count_hist[20];
#endif
};


extern struct evt_queue *
_evt_queue_init_internal( unsigned int depth );


extern struct evt_queue *
_evt_queue_init( unsigned int depth );


extern int
_evt_queue_deinit( exs_qhandle_t qhandle );


/* Change the wait type of the queue.  Can only be done prior to using the
 * queue for the first time. Called only from exs_qmodify().
 * Returns 0 if all ok, -1 and errno set on any error
 */
int
_set_evt_queue_wait(struct evt_queue *qptr, int *new_wait);

/* disables signal generation for this queue
 * returns 0 if all ok, -1 and errno set on any error
 */
extern int
_evt_queue_disable_signal( struct evt_queue *qhandle );


/* enables signal generation for this queue whenever a new event is
 * enqueued on a previously empty queue
 * returns 0 if all ok, -1 and errno set on any error
 */
extern int
_evt_queue_enable_signal( struct evt_queue *qhandle, int signal_number );


/* returns in vector up to count events posted to qhandle
 * returns the number actually copied into vector if all ok
 * returns -1 with errno set if any error
 *
 * FOR NOW, IGNORE TIMEOUT *** TO DO - fix this ***
 */
extern int
_evt_queue_dequeue_vector( struct evt_queue *qhandle, exs_event_t *vector,
			  int count, const struct timeval *timeout );


int
_get_empty_evt_queue_event(struct evt_queue *qptr, struct exs_event **evt_ptr);


extern int
_put_full_evt_queue_event( struct evt_queue *qhandle, exs_event_t *evt_ptr );


int
_put_empty_evt_queue_event(struct evt_queue *qptr, exs_event_t *evt_ptr);


/*
 * on entry, conn_pending->qhandle is known not to be NULL.
 * on exit, conn_pending->qhandle is always set to be NULL.
 *
 * if all ok, gets a new exs_evt_t item from qhandle's empty list,
 * fills it in with event info.
 *
 * returns 0 if all ok, else -1 with errno set (although result is never used!)
 */
extern int
_post_event_to_full_evt_queue(struct pending_connect *conn_pending);

void
_get_evt_queue_statistics(struct evt_queue *qptr,
					struct exs_qhandle_statistics *stats);

/* Called only from exs_qstatus()
 * Returns the current wait method for the given event queue
 */
void
_get_evt_queue_wait(struct evt_queue *qptr, int *result);

extern int
_get_evt_queue_depth( struct evt_queue *qhandle, int *result );


extern int
_get_evt_queue_signal( struct evt_queue *qhandle, struct exs_signal *sig);


extern int
_get_evt_queue_n_events( struct evt_queue *qhandle, int *result );

EXS_WARN_UNUSED_RETURN struct evt_queue *
_evt_queue_unref(struct evt_queue *qptr, unsigned int n, int flags);

extern struct evt_queue *
_cnv_qnumb_to_ptr(exs_qhandle_t qh, unsigned int n, int flags);

extern exs_qhandle_t
_cnv_ptr_to_qnumb( struct evt_queue *qh );

#endif /* _EVT_QUEUE_H_ */

/* vi: set autoindent tabstop=8 shiftwidth=8 : */
