/* libexs/trace.c */

/*
 * Copyright 2007-2016 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include <exs_private.h>

static struct timeval	base_tval;
static char		time_string[16];

#if defined(ALLOW_EXS_TRACE)
static struct trace_rule *our_ruleset = NULL;
static size_t our_ruleset_capacity = 0;
static size_t our_ruleset_size = 0;


static int
our_grow_ruleset(void)
{
	size_t new_capacity = 2 * our_ruleset_capacity;
	struct trace_rule *tmp;
	tmp = realloc(our_ruleset, new_capacity * sizeof(*our_ruleset));
	if (!tmp) {
		/* errno already set */
		return -1;
	}
	our_ruleset = tmp;
	our_ruleset_capacity = new_capacity;
	return 0;
}	/* our_grow_ruleset */


#define OUR_SKIP_WHITESPACE(input) \
	while (*(input) != '\0' && isspace(*(input))) \
		input++;


/* Scans a line range, which may be of the form:
 *
 *  line
 *  line1-line2
 *  line- (i.e., anywhere from line to EOF)
 *  -line (i.e., anywhere from beginning to line)
 *
 * On success, returns 0 and sets *endptr to the next input character.  On
 * failure, returns -1 and sets *endptr to the first input character. */
static int
our_scan_line_range(char *input, char **endptr, struct trace_rule *rule)
{
	char *ptr, *endp;
	long start = 1;
	long end = INT_MAX;

	OUR_SKIP_WHITESPACE(input);

	/* Parse start */
	if (*input == '-') {
		ptr = input + 1;
		if (*ptr == '\0') {
			*endptr = input;
			TRACE_DBG("Line range \"-\" has no numeric values");
			return -1;
		}
	} else {
		errno = 0;
		start = strtol(input, &ptr, 10);
		if (start < 1 || start > INT_MAX || errno != 0
							|| ptr == input) {
			TRACE_DBG("Invalid line range start in \"%s\"", input);
			*endptr = input;
			return -1;
		}
		if (*ptr == '-') {
			ptr++;
		} else if (*ptr != '\0' && !isspace(*ptr)) {
			*endptr = ptr;
			return -1;
		} else {
			end = start;
			endp = ptr;
		}
	}

	/* Parse end */
	if (*ptr != '\0' && !isspace(*ptr)) {
		errno = 0;
		end = strtol(ptr, &endp, 10);
		if (end < 0 || end > INT_MAX || errno != 0
					|| (*endp != '\0' && !isspace(*endp))) {
			TRACE_DBG("Invalid line range end in \"%s\"", input);
			*endptr = ptr;
			return -1;
		} else if (end < start) {
			TRACE_DBG("Line range \"%s\" is empty", input);
			*endptr = input;
			return -1;
		}
	}

	rule->line_min = start;
	rule->line_max = end;
	*endptr = endp;
	return 0;
}	/* our_scan_line_range */


static char *
our_strndup(const char *input, size_t n)
{
	char *output = malloc(n + 1);
	if (!output) {
		return NULL;
	}
	strncpy(output, input, n);
	output[n] = '\0';
	return output;
}	/* our_strndup */


static const size_t initial_output_string_capacity = 20;


/* Scan an octal escape sequence if base is 8, or a hexadecimal escape sequence
 * if base is 16.  The input string is expected to be advanced to the first
 * character of the numeric part of the escape sequence (i.e., to the first
 * digit of the octal sequence or one beyond the 'x' for a hexadecimal
 * sequence).  On success, returns the character read and sets *endptr to the
 * next input character.  On failure, returns the null character and sets
 * *endptr to the first input character. */
static char
our_scan_escape_sequence(const char *input, char **endptr, int base)
{
	long v;
	TRACE_DBG("scanning escaped base %d sequence at \"%s\"", base, input);
	errno = 0;
	v = strtol(input, endptr, base);
	if (v < CHAR_MIN || v > CHAR_MAX || errno != 0 || input == *endptr) {
		*endptr = (char *)input;
		return '\0';
	}
	return (char)v;
}	/* our_scan_escape_sequence */


/* Scan a C11 escape character sequence, including simple, octal, and
 * hexadecimal sequences.  The input string is expected to be advanced one
 * beyond the '\' that initiates the escape character sequence.  On success,
 * returns the character represented by the escape sequence and sets *endptr to
 * the next character in the input sequence.  On failure, returns the null
 * character and sets *endptr to the first input character. */
static char
our_scan_escaped_char(const char *input, char **endptr)
{
	char ch = *input;

	TRACE_DBG("scanning escaped char at \"%s\"", input);
	if (isdigit(ch)) {
		return our_scan_escape_sequence(input, endptr, 8);
	}
	switch (ch) {
	case 'x':
		return our_scan_escape_sequence(input + 1, endptr, 16);
	case '\0':
		*endptr = (char *)input;
		return ch;
	case 'a':
		ch = '\a';
		break;
	case 'b':
		ch = '\b';
		break;
	case 'f':
		ch = '\f';
		break;
	case 'n':
		ch = '\n';
		break;
	case 'r':
		ch = '\r';
		break;
	case 't':
		ch = '\t';
		break;
	case 'v':
		ch = '\v';
		break;
	default:
		/* '\'', '\"', '\\', and others represent themselves */
		break;
	}

	*endptr = (char *)input++;
	return ch;
}	/* our_scan_escaped_char */


static int
our_scan_char(const char *input, char **endptr, char **output,
		size_t *output_capacity, size_t *output_size)
{
	size_t cap = *output_capacity;
	size_t fill = *output_size;
	char *tmp;
	char nextch;

	if (*input != '\\') {
		nextch = *input++;
	} else {
		nextch = our_scan_escaped_char(input + 1, &tmp);
		if (!nextch) {
			TRACE_DBG("Invalid escape sequence at \"%s\"", input);
			*endptr = (char *)input;
			return -1;
		}
		input = tmp;
	}

	if (fill >= cap) {
		cap *= 2;
		tmp = realloc(*output, cap + 1);
		if (!tmp) {
			/* Caller will free output buffer */
			return -1;
		}
		*output_capacity = cap;
		*output = tmp;
	}
	(*output)[fill++] = nextch;
	*output_size = fill;
	*endptr = (char *)input;
	return 0;
}	/* our_scan_char */


static char *
our_scan_quoted_string(const char *input, char **endptr)
{
	const char *ptr;
	char *end;
	char *output;
	size_t cap;
	size_t fill;
	char mark;

	cap = initial_output_string_capacity;
	fill = 0;
	output = malloc(cap + 1);
	if (!output) {
		return NULL;
	}

	mark = *input;
	for (ptr = input + 1; *ptr != '\0' && *ptr != mark; ptr = end) {
		if (our_scan_char(ptr, &end, &output, &cap, &fill) < 0) {
			*endptr = (char *)ptr;
			free(output);
			return NULL;
		}
	}
	if (*ptr == '\0') {
		TRACE_DBG("Found unterminated string");
		*endptr = (char *)input;
		free(output);
		return NULL;
	}

	output[fill] = '\0';
	assert(*ptr == mark);
	ptr++;
	*endptr = (char *)ptr;
	return output;
}	/* our_scan_quoted_string */


static char *
our_scan_unquoted_string(const char *input, char **endptr)
{
	char *end;
	char *output;
	size_t cap;
	size_t fill;

	cap = initial_output_string_capacity;
	fill = 0;
	output = malloc(cap + 1);
	if (!output) {
		return NULL;
	}
	for (; *input != '\0' && !isspace(*input); input = end) {
		if (our_scan_char(input, &end, &output, &cap, &fill) < 0) {
			*endptr = (char *)input;
			free(output);
			return NULL;
		}
	}

	output[fill] = '\0';
	*endptr = (char *)input;
	return output;
}	/* our_scan_unquoted_string */


/* Scans a string, which consists of either: (1) a sequence of non-whitespace
 * characters, (2) a sequence of characters in single quotes, or (3) a sequence
 * of characters in double quotes.  Most escape sequences defined by C are
 * accepted in all three cases.  On success, returns a newly malloc'd buffer
 * containing the string (with each escape sequence replaced by the character
 * that the sequence represents) and sets *endptr to the next input character.
 * On failure, returns NULL and sets *endptr to the first input character. */
static char *
our_scan_string(const char *input, char **endptr)
{
	OUR_SKIP_WHITESPACE(input);
	if (*input == '\0') {
		return NULL;
	}

	switch (*input) {
	case '\'':
	case '\"':
		return our_scan_quoted_string(input, endptr);
	default:
		return our_scan_unquoted_string(input, endptr);
	}
}	/* our_scan_string */


/* Scans a keyword, which consists of letters.  On success, returns a newly
 * malloc'd buffer containing the keyword and sets *endptr to the next input
 * character.  On failure, returns NULL and sets *endptr to the first input
 * character. */
static char *
our_scan_keyword(const char *input, char **endptr)
{
	const char *end;

	OUR_SKIP_WHITESPACE(input);
	if (*input == '\0') {
		*endptr = (char *)input;
		return NULL;
	}

	end = input;
	while (*end != '\0' && isalpha(*end)) {
		end++;
	}

	*endptr = (char *)end;
	return our_strndup(input, end - input);
}	/* our_scan_keyword */


#ifndef EXPORT_PARSER
static
#endif
int
our_parse_rule(char *input, char **endptr, struct trace_rule *rule)
{
	char *token, *kwend, *end;
	int ret;

	memset(rule, 0, sizeof(*rule));
	end = input;
	token = our_scan_keyword(input, &kwend);
	if (token == NULL) {
		ret = -1;
		goto fail;
	}
	while (token != NULL) {
		if (!strcmp(token, "file")) {
			/* rule->file is guaranteed NULL or heap pointer */
			free(rule->file);
			if (!(rule->file = our_scan_string(kwend, &end))) {
				ret = -1;
				goto fail;
			}
			input = end;
		} else if (!strcmp(token, "line")) {
			if (our_scan_line_range(kwend, &end, rule) < 0) {
				ret = -1;
				goto fail;
			}
			input = end;
		} else if (!strcmp(token, "func")
				|| !strcmp(token, "function")) {
			/* rule->function is guaranteed NULL or heap pointer */
			free(rule->function);
			if (!(rule->function = our_scan_string(kwend, &end))) {
				ret = -1;
				goto fail;
			}
			input = end;
		} else if (!strcmp(token, "format")) {
			/* rule->format is guaranteed NULL or heap pointer */
			free(rule->format);
			if (!(rule->format = our_scan_string(kwend, &end))) {
				ret = -1;
				goto fail;
			}
			input = end;
		} else {
			/* Invalid token */
			ret = -1;
			goto fail;
		}
		free(token);
		token = our_scan_keyword(input, &kwend);
	}
	ret = 0;
fail:
	if (token) {
		free(token);
	}
	*endptr = end;
	return ret;
}	/* our_parse_rule */


static int
our_add_rule(struct trace_rule *rule)
{
	int index;

	if (!our_ruleset) {
		our_ruleset_capacity = 16;
		our_ruleset = EXS_CALLOC(our_ruleset_capacity,
				sizeof(*our_ruleset));
		if (!our_ruleset) {
			return -1;
		}
	}

	index = our_ruleset_size;
	if (our_ruleset_size >= our_ruleset_capacity) {
		if (our_grow_ruleset() != 0) {
			return -1;
		}
	}
	our_ruleset_size++;

	memcpy(&our_ruleset[index], rule, sizeof(*rule));
	TRACE_DBG("added rule %s %d-%d %s \'%s\'",
					rule->file, rule->line_min,
					rule->line_max, rule->function,
					rule->format);
	return 0;
}	/* our_add_rule */
#endif


/* store current time of day in global base_tval once at beginning of run */
static void
mark_time_base(void)
{
	gettimeofday(&base_tval, NULL);
	TRACE_DBG("%d.%06d: set base time",
			(int)base_tval.tv_sec, (int)base_tval.tv_usec);
}	/* mark_time_base */


#if defined(ALLOW_EXS_TRACE)
void
exs_trace_init(void)
{
	struct trace_rule rule;
	char line[BUFSIZ];
	char *end;
	char *envfile;
	FILE *ifptr;

	mark_time_base();

	if (!(envfile = getenv("EXS_TRACE_CONFIG"))) {
		our_ruleset = NULL;
		return;
	}

	ifptr = fopen(envfile, "r");
	if (!ifptr) {
		OUR_PERROR(errno, "Error opening %s for reading", envfile);
		TRACE_ERR("Using default trace config");
		return;
	}

	while (fgets(line, BUFSIZ, ifptr) != NULL) {
		if (our_parse_rule(line, &end, &rule) != 0) {
			if (*end != '\0') {
				TRACE_ERR("Ignoring trace rule with error at byte %td: %s",
						(end - line), line);
			}
		}
		if (our_add_rule(&rule) != 0) {
			OUR_PERROR(errno, "Error adding trace rule \"%s\"; giving up",
					line);
			break;
		}
	}
	if (fclose(ifptr) != 0) {
		OUR_PERROR(errno, "Error closing %s", envfile);
	}
}	/* exs_trace_init */
#else
void
exs_trace_init(void)
{
	mark_time_base();
}	/* exs_trace_init */
#endif


/* create a time-stamp for debug trace printing in global time_string array */
static char *
_time_prefix(void)
{
	struct timeval	cur_tval;

	gettimeofday(&cur_tval, NULL);
	if((int)base_tval.tv_usec > (int)cur_tval.tv_usec) {
		cur_tval.tv_usec += 1000000;
		cur_tval.tv_sec -= 1;
	}
	cur_tval.tv_usec -= base_tval.tv_usec;
	cur_tval.tv_sec -= base_tval.tv_sec;
	sprintf(time_string, "%3d.%06d:", (int)cur_tval.tv_sec,
		(int)cur_tval.tv_usec);
	return time_string;
}	/* _time_prefix */


/* the number of bytes to allocate for strerror_r() buffer */
#define	OUR_BUF_SIZE	52


/* Prints a prefix of the form:
 * "<timestamp> <file>:<line> <function> [<threadid>]"
 */
static inline void
our_log_prefix(FILE *fptr, const char *file, int line, const char *function)
{
	pthread_t pth_id = pthread_self();
	unsigned long long threadid;

	/* POSIX doesn't say anything about the structure of pthread_t, so
	 * there's no portable way to print it.  The closest thing we can do is
	 * to cast it to an appropriately-sized integer and hope for the
	 * best. */
	if (sizeof(pth_id) == sizeof(uint8_t)) {
		threadid = *(uint8_t *)&pth_id;
	} else if (sizeof(pth_id) < sizeof(uint32_t)) {
		threadid = *(uint16_t *)&pth_id;
	} else if (sizeof(pth_id) < sizeof(uint64_t)) {
		threadid = *(uint32_t *)&pth_id;
	} else {
		threadid = *(uint64_t *)&pth_id;
	}

	fprintf(fptr, "%s %s:%d %s [%llu] ",
			_time_prefix(), file, line, function,
			threadid);
}	/* our_log_prefix */


static bool
our_match_rules(const char *file, int line, const char *function,
		const char *format)
{
#if !defined(ALLOW_EXS_TRACE)
	(void)file;
	(void)line;
	(void)function;
	(void)format;
#else
	struct trace_rule *r;
	unsigned int i;

	for (i = 0; i < our_ruleset_size; i++) {
		r = &our_ruleset[i];
		if (r->file && strcmp(r->file, file) != 0) {
			continue;
		} else if (r->line_min && line < r->line_min) {
			continue;
		} else if (r->line_max && line > r->line_max) {
			continue;
		} else if (r->function && strcmp(r->function, function) != 0) {
			continue;
		} else if (r->format && strstr(format, r->format) == NULL) {
			continue;
		} else {
			/* We found a match */
			return true;
		}
	}
#endif
	return false;
}	/* our_match_rules */


/* Prints a message of the following form:
 *
 * "<our_log_prefix output> <formatted_msg>\n"
 *
 * This function will preserve the value of errno, so it is safe to call on
 * error paths. */
void
exs_log(int priority, const char *file, int line, const char *function,
		const char *format, ...)
{
	FILE *fptr;
	va_list args;
	int old_errno = errno;
	(void)priority;

	if (priority == EXS_LOG_DEBUG && !TRACING_IS_ON
			&& !our_match_rules(file, line, function, format)) {
		return;
	}

	fptr = stderr; /* to get around git pre-commit hook */
	flockfile(stderr);
	our_log_prefix(fptr, file, line, function);
	va_start(args, format);
	vfprintf(fptr, format, args);
	va_end(args);
	putc_unlocked('\n', stderr);
	funlockfile(stderr);
	errno = old_errno;
}	/* exs_log */


/* Prints a message of the following form:
 *
 * "<our_log_prefix output> <formatted_msg>: <strerror(errnum)>\n"
 *
 * except that this function calls the appropriate thread-safe version of
 * strerror() for this system using a stack-allocated buffer.  Additionally,
 * this function will preserve the value of errno, regardless of the source of
 * the errnum argument, so it is safe to call on error paths. */
void
exs_perror(const char *file, int line, const char *function,
		int errnum, const char *format, ...)
{
	/* This is the easiest way to do things for now, but there is a very
	 * small probability that this could cause the stack to overflow.  Since
	 * this function should never be called anyways, we will accept this
	 * risk for now. */
	char errbuf[OUR_BUF_SIZE];

	FILE *fptr;
	va_list args;
	int old_errno = errno;

	if (strerror_r(errnum, errbuf, OUR_BUF_SIZE)) {
		snprintf(errbuf, OUR_BUF_SIZE, "Unknown error %d", errnum);
	}

	fptr = stderr; /* to get around git pre-commit hook */
	flockfile(stderr);
	our_log_prefix(fptr, file, line, function);
	va_start(args, format);
	vfprintf(fptr, format, args);
	va_end(args);
	fprintf(fptr, ": %s\n", errbuf);
	funlockfile(stderr);
	errno = old_errno;
}	/* exs_perror */


#if defined(ALLOW_EXS_TRACE)

void
our_dump_buf(FILE *fptr, void *buffer, int length)
{
	int	i, actual_length;
	char	*ptr, dumpbuf[120];

		/* dump only min(length, 500) bytes */
		actual_length = length;
		if(length > 500)
			length = 500;
		/* dump the buffer in hex, 16 bytes per line */
		ptr = dumpbuf + sprintf(dumpbuf, "%11d:", 0);;
		for(i = 0; i < length; i++)
		{
			ptr = ptr + sprintf(ptr, " %02x",
					*(((unsigned char *)buffer) + i));
			if(((i + 1) & 0x0f) == 0)
			{
			fprintf(fptr, "%s\n", dumpbuf);
			ptr = dumpbuf + sprintf(dumpbuf, "%11d:", i+1);;
			}
		}
		fprintf(fptr, "%s\n", dumpbuf);
		if(length < actual_length)
			fprintf(fptr, "%12s %d remaining bytes not dumped\n",
				"", actual_length - length);
}	/* our_dump_buf */

#endif
